/********************************************************************
 ** @version BRA10
 ** P: Support pointer data
 ** @author Alvaro Estebanez
 ** @date 1th March 2012
 *********************************************************************/

//  Main header file with all common variables
#include "appcommon.h"
#include "appcommon2.h"
#include "constants.h"
//  speccode: OpenMP header file included
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
//#include <signal.h>



//***********************************************************************
//   specstart(): operation to initialize speculation before every
//                speculative section
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************

int specstart(int iterations) {
    //  Initialize the wheel structure
    wheel_ns = 0;
    wheel_ms = -1;
    int thread;
    /******************************************************************************/
    // Initialize wheel structure for each position
    for (thread = 0; thread < wsize; thread++) {
        wheel[thread] = FREE;

        wheel_data[thread].pointer = (threadCopy_structure *) malloc(sizeof (threadCopy_structure));
#ifdef SPECSTART_DEBUG
    printf("specstart: wheel_data[thread].pointer=%p\n",wheel_data[thread].pointer);
    printf("specstart: size of requested data:%i\n",sizeof (threadCopy_structure));
    fflush(stdout);
#endif
        
        wheel_data[thread].lastData = 0;
        wheel_data[thread].maxData = MAX_POINTER;
    }

    /**
     **  initialize the number of iterations to execute in the loop
     */
    upper_limit = iterations;

    /******************************************************************************/
    //  Initialize the partial loop bounds
    lower_bound = 1;
    upper_bound = wsize*blk;

    return (EXIT_SUCCESS);
}

/**
 ** @version specload-P
 ** @author alvaro estebanez
 **/
//*********************************************************************
//   specload_pointer(): speculative load operation with forwarding
//*********************************************************************
/**
 *
 * @param address   - Address of the pointer to be loaded
 * @param byteSize  - Size of the data to be loaded
 * @param current   - The number of thread
 * @param value     - Address of the variable which loads data pointed
 * @return          - EXIT_SUCCESS if everything is right
 **/
int specload_pointer(unsigned char* address, unsigned int byteSize,
        unsigned int current, unsigned char* value) {

    // Check if thread is squashed
#pragma omp flush(wheel)
    if (wheel[modvector[current]] == SQUASHED) {
        return -1;
    }

#ifdef SPECLOAD_DEBUG
//  speccode: debug
  tid = omp_get_thread_num();
  printf("specload: tid = %d, current = %d modvector(current) = %d,..Beginning load\n",omp_get_thread_num(),current,modvector[current]);
  printf("specload_pointer: byteSize = %i\n",byteSize);
  fflush(stdout);
//  speccode: end debug
#endif

    // local variables
    int punteroUtilizado = false;
    int thread = current;
    int row = 0;
    int lastData = 0;
#pragma omp flush(wheel_data)
    threadCopy_structure *punteroMiEstructura = wheel_data[modvector[thread]].pointer;
#ifdef SPECLOAD_DEBUG
    printf("specload_pointer: End first assignment. punteroMiEstructura=%p\n",punteroMiEstructura);
    printf("specload_pointer: Beginning second assignment...\n");
    fflush(stdout);
#endif
    threadCopy_structure miEstructura = *punteroMiEstructura;
#ifdef SPECLOAD_DEBUG
    printf("specload_pointer: End second assignment.\n");
    fflush(stdout);
#endif

    threadCopy_structure *punteroEstructuraAux = NULL;
    threadCopy_structure estructuraAux;
#ifdef SPECLOAD_DEBUG
    printf("specload_pointer: wheel_ns = %i\n",wheel_ns);
    printf("specload_pointer: wheel_ms = %i\n",wheel_ms);
    fflush (stdout);
#endif

    // check if the pointer to be loaded has been used by some thread
    while ((!punteroUtilizado) && (thread >= wheel_ns)) {
        if ( (miEstructura.matriz[T_POINTER][row] == (baseType) address) && (wheel_data[modvector[thread]].lastData != row) ) {
            punteroUtilizado = true;
            // thread variable will contain the thread which used the pointer
            // row variable will contain the position of the pointer in its matrix
        } else {
            // Check if get last position of matrix
            //#pragma omp flush(wheel_data)
            if (row < wheel_data[modvector[thread]].lastData)
                row++;
            else { // If get last position, must looking for other thread
                thread--;
                row = 0;
                if (thread >= wheel_ns) {
                    #pragma omp flush(wheel_data)
                    punteroMiEstructura = wheel_data[modvector[thread]].pointer;
                    miEstructura = *punteroMiEstructura;
                }
            }
        }
    }

    // NOTE: thread variable will contain the thread which used the pointer
    //       row variable will contain the position of the pointer in its matrix
    //       If the threads dont use the pointer:
    //          thread = wheel_ns - 1;
    //          row = 0;
    //          punteroUtilizado = false;
#pragma omp flush(wheel_data)
    punteroMiEstructura = wheel_data[modvector[current]].pointer;
    miEstructura = *punteroMiEstructura;
    lastData = wheel_data[modvector[current]].lastData;

    unsigned char *apuntador = NULL;
    unsigned char *apuntador2 = NULL;
    int i = 0;
#ifdef SPECLOAD_DEBUG
    printf("specload_pointer: Ya se ha buscado el dato.\n");
    fflush (stdout);
#endif

    // If the thread that wants to load data had used pointer before then
    // must load it from its matrix
    if (current == thread) {
        if (miEstructura.matriz[T_STATE][row] == UPDATE) {
             //|| (miEstructura.matriz[T_STATE][row] == ELUP) ) {
            miEstructura.matriz[T_STATE][row] = ELUP;
        }/* else {
            miEstructura.matriz[T_STATE][row] = EXPLD;
        }*/

        // Checking for data size
        apuntador = (unsigned char *) miEstructura.matriz[T_DATA][row];
        switch (miEstructura.matriz[T_SIZE][row]) {
            case 1:
                value[0] = apuntador[0];
                break;
            case 2:
                value[0] = apuntador[0];
                value[1] = apuntador[1];
                break;
            case 4:
                value[0] = apuntador[0];
                value[1] = apuntador[1];
                value[2] = apuntador[2];
                value[3] = apuntador[3];
                break;
            case 8:
                value[0] = apuntador[0];
                value[1] = apuntador[1];
                value[2] = apuntador[2];
                value[3] = apuntador[3];
                value[4] = apuntador[4];
                value[5] = apuntador[5];
                value[6] = apuntador[6];
                value[7] = apuntador[7];
                break;
            default:
                for (i = 0; i < miEstructura.matriz[T_SIZE][row]; i++) {
                    value[i] = apuntador[i];
                }
                break;
        }
    } else { // current != thread
        #pragma omp flush(wheel_data)
        lastData = wheel_data[modvector[current]].lastData;
        // Check if thread matrix is full (need to add data)
        //if (wheel_data[modvector[current]].lastData == wheel_data[modvector[current]].maxData)
        //    reservarMasEspacio();
        miEstructura.matriz[T_STATE][lastData] = EXPLD;
        miEstructura.matriz[T_SIZE][lastData] = byteSize;
        miEstructura.matriz[T_POINTER][lastData] = (baseType) address;
        apuntador = (unsigned char *) malloc(byteSize);
#ifdef SPECLOAD_DEBUG
    printf("specload: apuntador=%p\n",apuntador);
    printf("specload: size of requested data:%i\n", byteSize);
    fflush(stdout);
#endif
        if (punteroUtilizado) { // If some thread used pointer, must load it from its matrix
            #pragma omp flush(wheel_data)
            punteroEstructuraAux = wheel_data[modvector[thread]].pointer;
            estructuraAux = *punteroEstructuraAux;
            // miEstructura     ->  belongs to currentThread
            // estructuraAux    ->  belongs to the thread which used data
            //miEstructura.matriz[T_POINTER][lastData] = estructuraAux.matriz[T_POINTER][row];
            //miEstructura.matriz[T_SIZE][lastData] = estructuraAux.matriz[T_SIZE][row];
            
            // apuntador variables are needed because they may save values byte to byte
            //apuntador = (unsigned char *) malloc (estructuraAux.matriz[T_SIZE][row]);
            apuntador2 = (unsigned char *) estructuraAux.matriz[T_DATA][row];
        } else {
            // No thread used the pointer so load value from memory
            //miEstructura.matriz[T_POINTER][lastData] = (baseType) address;
            //miEstructura.matriz[T_SIZE][lastData] = byteSize;
            //apuntador = (unsigned char *) malloc (byteSize);
            apuntador2 = address;
        }
        // apuntador    -> contain a direction to save data
        // apuntador2   -> contain the most recently version of the pointer
        // Checking for data size
        switch (byteSize) {
            case 1:
                apuntador[0] = apuntador2[0];
                value[0] = apuntador[0];
                break;
            case 2:
                apuntador[0] = apuntador2[0];
                apuntador[1] = apuntador2[1];
                value[0] = apuntador[0];
                value[1] = apuntador[1];
                break;
            case 4:
                apuntador[0] = apuntador2[0];
                apuntador[1] = apuntador2[1];
                apuntador[2] = apuntador2[2];
                apuntador[3] = apuntador2[3];
                value[0] = apuntador[0];
                value[1] = apuntador[1];
                value[2] = apuntador[2];
                value[3] = apuntador[3];
                break;
            case 8:
                apuntador[0] = apuntador2[0];
                apuntador[1] = apuntador2[1];
                apuntador[2] = apuntador2[2];
                apuntador[3] = apuntador2[3];
                apuntador[4] = apuntador2[4];
                apuntador[5] = apuntador2[5];
                apuntador[6] = apuntador2[6];
                apuntador[7] = apuntador2[7];
                value[0] = apuntador[0];
                value[1] = apuntador[1];
                value[2] = apuntador[2];
                value[3] = apuntador[3];
                value[4] = apuntador[4];
                value[5] = apuntador[5];
                value[6] = apuntador[6];
                value[7] = apuntador[7];
                break;
            default:
                for (i = 0; i < byteSize; i++) {
                    apuntador[i] = apuntador2[i];
                    value[i] = apuntador[i];
                }
                break;
        }
        miEstructura.matriz[T_DATA][lastData] = (baseType) apuntador;
        // Data added so must increase number of data in the matrix
        wheel_data[modvector[current]].lastData++;
        //#pragma omp flush(wheel_data)
    }

    // Asign the structure with the changes that has been made
    *(wheel_data[modvector[current]].pointer) = miEstructura;
#pragma omp flush(wheel_data)

    // Free pointers
    punteroMiEstructura = NULL;
    punteroEstructuraAux = NULL;
    apuntador = NULL;
    apuntador2 = NULL;

    return EXIT_SUCCESS;

}// END specload_pointer

/**
 ** @version specstore-P
 ** @author alvaro estebanez
 **/
//*********************************************************************
//   specstore_pointer(): speculative store operation
//*********************************************************************
/**
 *
 * @param address   - Address of the pointer to be saved
 * @param byteSize  - Size of the data to be saved
 * @param current   - The number of thread
 * @param value     - Address of the variable which contain the value to be saved
 * @return          - EXIT_SUCCESS if everything is right
 **/
int specstore_pointer(unsigned char* address, unsigned int byteSize,
        unsigned int current, unsigned char* value) {

    // Check if thread is squashed
    #pragma omp flush(wheel)
    if (wheel[modvector[current]] == SQUASHED) {
#ifdef SPECSTORE_DEBUG
//  speccode: debug
  printf("ESTOY SQUASHED!.specstore: tid = %d, current = %d,\n",omp_get_thread_num(),current);
  fflush(stdout);
//  speccode: end debug
#endif
        return -1;
    }

#ifdef SPECSTORE_DEBUG
//  speccode: debug
  printf("specstore: tid = %d, current = %d, address=0x%x Beginning store\n",omp_get_thread_num(),current,address);
  fflush(stdout);
//  speccode: end debug
#endif

    // local variables
    int punteroUtilizado = false;
    int thread = current;
    int row = 0;
    int lastData = 0;
    #pragma omp flush(wheel_data)
    threadCopy_structure *punteroMiEstructura = wheel_data[modvector[thread]].pointer;
    threadCopy_structure miEstructura = *punteroMiEstructura;
    lastData = wheel_data[modvector[current]].lastData;


    threadCopy_structure *punteroEstructuraAux = NULL;
    threadCopy_structure estructuraAux;

    #pragma omp flush(wheel_data)
    // check if the direction where I want to save data, has been used by current thread
    //while ((!punteroUtilizado) && (row < wheel_data[modvector[current]].lastData)) {
    while ((!punteroUtilizado) && (row < lastData)) {
        if (miEstructura.matriz[T_POINTER][row] == (baseType) address) {
            punteroUtilizado = true;
            // row variable will contain the position of the pointer in its matrix
        } else {
            row++;
        }
    }
    // NOTE: row variable will contain the position of the pointer in its matrix
    unsigned char *apuntador = NULL;
    int i = 0;

    // if current thread had used this value before then must override value
    // from its matrix
    if (punteroUtilizado) {
        // First of all, store the value
        // Checking for data size
        apuntador = (unsigned char *) miEstructura.matriz[T_DATA][row];
        switch (miEstructura.matriz[T_SIZE][row]) {
            case 1:
                apuntador[0] = value[0];
                break;
            case 2:
                apuntador[0] = value[0];
                apuntador[1] = value[1];
                break;
            case 4:
                apuntador[0] = value[0];
                apuntador[1] = value[1];
                apuntador[2] = value[2];
                apuntador[3] = value[3];
                break;
            case 8:
                apuntador[0] = value[0];
                apuntador[1] = value[1];
                apuntador[2] = value[2];
                apuntador[3] = value[3];
                apuntador[4] = value[4];
                apuntador[5] = value[5];
                apuntador[6] = value[6];
                apuntador[7] = value[7];
                break;
            default:
                for (i = 0; i < miEstructura.matriz[T_SIZE][row]; i++) {
                    apuntador[i] = value[i];
                }
                break;
        }
//#pragma omp flush(miEstructura)

        miEstructura.matriz[T_DATA][row] = (baseType) apuntador;

        if (miEstructura.matriz[T_STATE][row] == EXPLD) {
            miEstructura.matriz[T_STATE][row] = ELUP;
        }// else {
         //   miEstructura.matriz[T_STATE][row] = UPDATE;
        //}
//#pragma omp flush(miEstructura)
        
    } else { // current thread have not used the pointer
#pragma omp flush(wheel_data)
        lastData = wheel_data[modvector[current]].lastData;
        // Check if thread matrix is full (need to add data)
        //if (wheel_data[modvector[current]].lastData == wheel_data[modvector[current]].maxData)
        //    reservarMasEspacio();

        // First of all, store the value
        // apuntador variables are needed because it is need to save values byte to byte
        apuntador = (unsigned char *) malloc (byteSize);
        // Checking for data size
        switch (byteSize) {
            case 1:
                apuntador[0] = value[0];
                break;
            case 2:
                apuntador[0] = value[0];
                apuntador[1] = value[1];
                break;
            case 4:
                apuntador[0] = value[0];
                apuntador[1] = value[1];
                apuntador[2] = value[2];
                apuntador[3] = value[3];
                break;
            case 8:
                apuntador[0] = value[0];
                apuntador[1] = value[1];
                apuntador[2] = value[2];
                apuntador[3] = value[3];
                apuntador[4] = value[4];
                apuntador[5] = value[5];
                apuntador[6] = value[6];
                apuntador[7] = value[7];
                break;
            default:
                for (i = 0; i < byteSize; i++) {
                    apuntador[i] = value[i];
                }
                break;
        }
//#pragma omp flush(miEstructura)

        miEstructura.matriz[T_STATE][lastData] = UPDATE;
        miEstructura.matriz[T_POINTER][lastData] = (baseType) address;
        miEstructura.matriz[T_SIZE][lastData] = byteSize;
//#pragma omp flush(miEstructura)
        
        miEstructura.matriz[T_DATA][lastData] = (baseType) apuntador;
        wheel_data[modvector[current]].lastData++;
        //#pragma omp flush(wheel_data)
    }

    // Asign the structure with the changes that has been made
    *(wheel_data[modvector[current]].pointer) = miEstructura;
#pragma omp flush(wheel_data)

    // Now, check that does not exist any succesor thread that had used the
    // pointer, if this situation occurs then thread execution is discarded
    int specaux1, squa, aux;
    char localsquash;
    aux = modvector[current];

//  Look for possible RAW dependences.
//  New since version 140: RAW dependences should be searched
//  even if the thread is ms, since this situation could
//  change during execution of this loop.

    localsquash = 'F';

#ifdef SPECSTORE_DEBUG
//  speccode: debug
  printf("Checking for squashes between %d and %d. SOY:current=%d,tid=%d.\n",current+1,wheel_ms,current,omp_get_thread_num());
  for (specaux1=current+1;specaux1<=wheel_ms;specaux1++)
      printf("\tESTADO(%d):%d\n",specaux1,wheel[specaux1]);
  fflush(stdout);
//  speccode: end debug
#endif
    specaux1 = current + 1;
e1122:
//    El compilador no deja poner etiquetas justo antes de una instruccion omp
      if(1)
      {
        #pragma omp flush(wheel_ms)
          }
      if ( specaux1 > wheel_ms )
        goto e2233;
      else
      {
        #pragma omp flush(wheel_data)
        punteroEstructuraAux = wheel_data[modvector[specaux1]].pointer;
        estructuraAux = *punteroEstructuraAux;
        row = 0;
        while ( (localsquash == 'F') && (row < wheel_data[modvector[specaux1]].lastData) ) {
        //for (row = 0; row < wheel_data[modvector[specaux1]].lastData; row++) {
            if (estructuraAux.matriz[T_POINTER][row] == (baseType) address) {
                // Pointer exists, so it has been used and must be discarded
                if ( estructuraAux.matriz[T_STATE][row] == UPDATE ) {
#ifdef SPECSTORE_DEBUG
//  speccode: debug
  printf("specstore: tid = %d, current = %d Shield found. Exiting DO loop\n",tid,current);
  fflush(stdout);
                    //  speccode: end debug
#endif
                    //specaux1 = specaux1 + 1;
                    //goto e1122;
                    goto e2233;
                }
                if ( estructuraAux.matriz[T_STATE][row] != NOTACC ) {
        //      SQUASH!

        //***********************************************************************

                  #pragma omp critical
                  {
                    // Statistics
                    SC_specstore++;

#ifdef SPECSTORE_DEBUG
//  speccode: debug
  printf("specstore: tid = %d, current = %d, estado = %d\n",tid,current,wheel[aux]);
  fflush(stdout);
//  speccode: end debug
#endif
#pragma omp flush(wheel)
                    if ( ( wheel[aux] == RUNNING ) && ( specaux1 <= wheel_ms ) )
                    {
        //          begin squash critical section

        //          We increment the spec_violations counter here
                      spec_violations = spec_violations + 1;

        //          Squash "specaux1" and all its successors.
        //          If a sucessor is RUNNING, this is set as SQUASHED.
        //          If a sucessor is DONE, then is set as FREE.

#pragma omp flush(wheel_ms)

                      for ( squa = specaux1 ; squa <= wheel_ms ; squa++ )
                      {
        //            We increment the number of squashes performed.
                        spec_squashes = spec_squashes + 1;

#pragma omp flush(wheel)
                        if ( wheel[modvector[squa]] == RUNNING ) {
#ifdef SPECSTORE_DEBUG
// speccode: debug
   printf("Consumer %d RUNNING->SQUASHED, tid = %d\n", squa, omp_get_thread_num());
   fflush(stdout);
// speccode: end debug
#endif

                            wheel[modvector[squa]] = SQUASHED;
                            wheel_data[modvector[squa]].lastData = 0;

                        }

                        if ( wheel[modvector[squa]] == DONE ) {
#ifdef SPECSTORE_DEBUG
// speccode: debug
   printf("Consumer %d DONE->FREE, tid = %d\n", squa, omp_get_thread_num());
   fflush(stdout);
// speccode: end debug
#endif
                                wheel[modvector[squa]] = FREE;
                                wheel_data[modvector[squa]].lastData = 0;

                        }

#pragma omp flush(wheel, wheel_data, wheel_ms)
                      } // END FOR
#ifdef SPECSTORE_DEBUG
// speccode: debug
   printf("Last consumer squashed= %d\n", wheel_ms);
   fflush(stdout);
// speccode: end debug
#endif

#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif

                      wheel_ms = specaux1 - 1;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif
#pragma omp flush(wheel_ms, wheel_ns)

#ifdef SPECSTORE_DEBUG
// speccode: debug
   printf("New most-spec thread= %d, Actual non-spec=%d.\n", wheel_ms, wheel_ns);
   fflush(stdout);
// speccode: end debug
#endif
                      localsquash = 'T';


                    } //END IF((wheel[tidaux]==RUNNING)&&(specaux1<=wheel_ms))
#ifdef SPECSTORE_DEBUG
                    else {
// speccode: debug
   printf("NO ESTOY RUNNING, SOY %d.\n", omp_get_thread_num());
   fflush(stdout);
// speccode: end debug
                    }
#endif


                  } // END OMP CRITICAL

        //***********************************************************************

        // speccode: end squash critical section.

                } // END IF(AV[index,modvector[specaux1]] != NOTACC)
                //else
                  // PARA PONER UN DEBUG
            } // end if (estructuraAux.matriz[T_POINTER][row] == (baseType) address)
            row++;
        //} // end for (row = 0; row < wheel_data[modvector[specaux1]].lastData; row++)
        }//while ( (localsquash == 'F') && (row < wheel_data[modvector[specaux1]].lastData) )
        if (localsquash == 'T')
        {
//      The squashes has been performed.
//      We should exit the DO loop.
          goto e2233;
        }


      } // END IF(specaux1 >= wheel_ms)

//    v140 :  Now, we try to repeat the trick.
      specaux1 = specaux1 + 1;
      goto e1122;

//    Once the searche is finished, label "e2233" is reached.
e2233: ;
//   continue;

    // Asign the structure with the changes that has been made
    //*(wheel_data[modvector[current]].pointer) = miEstructura;
    //#pragma omp flush(wheel_data)

    // Free pointers
    punteroMiEstructura = NULL;
    punteroEstructuraAux = NULL;
    apuntador = NULL;

    return EXIT_SUCCESS;

}// END specstore_pointer

/**
 ** @version specstore-P
 ** @author alvaro estebanez
 **/
//***********************************************************************
//   specinit(): operation to initialize blocks
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************
void specinit() {

    //  real*8   IEEE double-precision floating-point (Eight-byte)
    int specaux1;

    //  Initialize the modvector structure (only once).

    upper_limit = cur_upper_limit;

    for (specaux1 = 0; specaux1 <= (max_upper_limit + 8 - 1); specaux1++)
        modvector[specaux1] = ((specaux1) % (wsize));

    //  Depending on the value of blk, generate a fixed-block scheduling
    //  or read a variable one from disk.

    if (blk > 0) {
        //  Fixed block size: initialize vector with it.
        //  upper_limit & blk al ser enteros la division es entera
        for (maxblock = 0; maxblock <= (upper_limit / blk); maxblock++) {

            //    First element: iteration to start
            varblock[0][maxblock] = blk * (maxblock);

            //    Second element: iteration to finish
            varblock[1][maxblock] = blk * (maxblock + 1) - 1;

            //    Third element: Number of attempts for this block
            varblock[2][maxblock] = 0;
#ifdef SPECINIT_DEBUG
            printf("New block: from %d to %d\n",varblock[0][maxblock],varblock[1][maxblock]);
#endif
        }

    }
} // end specinit()

/**
 ** @version specstore-P
 ** @author alvaro estebanez
 ****************************************************************************
 ** @param - current block number
 ** @param - nonSpeculative flag to indicate if the thread is the non speculative
 ** @return - 0: it have not any block left
              1: new block assigned
 */
int getFirstBlock(int *current, int *nonSpeculative) {

    int retorno = 0;
#ifdef GETFIRSTBLOCK_DEBUG
//  speccode: debug
    printf("getFirstBlock: Hola soy: %d , numero de threads %d\n",omp_get_thread_num(),omp_get_num_threads());
    fflush(stdout);
//  speccode: end_debug
#endif

#pragma omp critical
    {
        // There is no more work to do
#pragma omp flush(wheel_ms)
        if (varblock[0][wheel_ms + 1] > upper_limit - 1)
            retorno = 0;
        else {
            // Only if the ms pointer plus one is pointing to a
            // FREE cell we can proceed
#pragma omp flush(wheel, wheel_ms)
#ifdef GETFIRSTBLOCK_DEBUG
    printf("getFirstBlock(): Aun queda trabajo.\n");
    fflush(stdout);
#endif

            if (wheel[modvector[wheel_ms + 1]] == FREE) {
                // Clean the access vector of the cell to be used
                wheel_data[modvector[wheel_ms + 1]].lastData = 0;
                // Modify the status of the window cell to be used
                wheel[modvector[wheel_ms + 1]] = RUNNING;

#pragma omp flush(wheel, wheel_ms, wheel_data)
                // Advance the wheel_ms pointer
#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif

                wheel_ms = wheel_ms + 1;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif
                #pragma omp flush(wheel_ms, wheel_ns)
                *nonSpeculative = (wheel_ms == wheel_ns);

#pragma omp flush(wheel_ms)

                // Prepare the chunk for running
                *current = wheel_ms;
                retorno = 1;

            } // END if ( wheel[wheel_ms+1] == FREE )
        } // END else ( varblock[1,wheel_ms] >= upper_limit )
    }
#ifdef GETFIRSTBLOCK_DEBUG
    printf("getFirstBlock(): retorno vale %i\n",retorno);
    fflush(stdout);
#endif
    return retorno;
    //***********************************************************************
}

/**
 ** @version specstore-P
 ** @author alvaro estebanez
 **/
//*********************************************************************
//   threadend_pointer():   this operation is launch when a thread finish
//                          its block
//*********************************************************************
/**
 *
 * @param current           - Position of the thread
 * @param nonSpeculative    - To know if there are only one thread running
 * @return                  - Notify if there are some job to do
 **/
char threadend_pointer(int *current, int *nonSpeculative) {

    //  Local work flags
    char auxdone;
    char auxjobtodo;
    int aux;
    char flag;
#ifdef THREADEND_DEBUG
//  speccode: debug
  printf("threadend: tid = %d, current = %d Beginning thread end\n",omp_get_thread_num(),*current);
  fflush(stdout);
//  speccode: end debug
#endif

    //  Set work flags
    auxdone = 'F';
    auxjobtodo = 'F';
    flag = JOBTODO;

    //***********************************************************************
#pragma omp critical
    {
        //  Statistics
        SC_threadend1++;

#pragma omp flush(wheel)

        if (wheel[modvector[*current]] != SQUASHED) {
#pragma omp flush(wheel_ns, wheel_ms)

#ifdef THREADEND_DEBUG
//  speccode: debug
  printf("threadend: tid= %d, current= %d I was not squashed\n",omp_get_thread_num(), *current);
  fflush(stdout);
//  speccode: end debug
#endif
            // Now check if non-speculative
            if (*current == wheel_ns) {
                // No-speculative thread finished
#ifdef THREADEND_DEBUG
//  speccode: debug
  printf("threadend: tid= %d, current= %d I am ns! Calling commit()\n",omp_get_thread_num(), *current);
  fflush(stdout);
//  speccode: end debug
#endif
                aux = commit();
#ifdef WHEEL_NS_DEBUG
    printf("cambio en wheel_ns: ANTES=%d,",wheel_ns);
    fflush(stdout);
#endif
                wheel_ns = aux + 1;
#ifdef WHEEL_NS_DEBUG
    printf(" DESPUES=%d \n",wheel_ns);
    fflush(stdout);
#endif
                #pragma omp flush(wheel_ns)
#ifdef THREADEND_DEBUG
  printf("TID: %d, Actualizando wheel_ns: %d\n",omp_get_thread_num(),wheel_ns);
  fflush(stdout);
#endif

                if (varblock[1][aux] >= upper_limit - 1) {
                    // Mark this thread as done
                    auxdone = 'T';
                } else {
                    // Move the non-speculative pointer
#pragma omp flush(wheel_ms, wheel_ns)

                    // If the block being executed by the most speculative
                    // thread has an upper iteration that is greater than or
                    // equal to the last iteration to perform, then we finish
                    if (varblock[1][wheel_ms] >= upper_limit - 1)
                        auxdone = 'T';
                    else {
                        // Only if the ms pointer plus one is pointing to a
                        // FREE cell we can proceed
#pragma omp flush(wheel, wheel_ms)

                        if (wheel[modvector[wheel_ms + 1]] == FREE) {
                            // Clean the access vector of the cell to be used
                            wheel_data[modvector[wheel_ms + 1]].lastData = 0;
                            // Modify the status of the window cell to be used
                            wheel[modvector[wheel_ms + 1]] = RUNNING;

#pragma omp flush(wheel, wheel_ms, wheel_data)
                            // Advance the wheel_ms pointer
#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif
                            wheel_ms = wheel_ms + 1;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif


                            /**  nonSpeculative will be 1 if you are the thread
                            **  non speculative before starting the job. Otherwise
                            **  the value will be 0 **/
                            #pragma omp flush(wheel_ns, wheel_ms)
#ifdef THREADEND_DEBUG
  printf("TID: %d, Actualizando wheel_ms: %d\n",omp_get_thread_num(),wheel_ms);
  fflush(stdout);
#endif
                            *nonSpeculative = (wheel_ms == wheel_ns);

#pragma omp flush(wheel_ms)

                            // Prepare the chunk for running
                            auxjobtodo = 'T';
                            *current = wheel_ms;

                        } // END if ( wheel[modvector[wheel_ms+1]] == FREE )
                    } // END else ( varblock[1,wheel_ms] >= upper_limit )
                } // END else ( varblock[1,tidaux] >= upper_limit )
            }// END if ( *current == wheel_ns )
            else {

                // Speculative thread finished
#ifdef THREADEND_DEBUG
//  speccode: debug
  printf("threadend: tid= %d, current= %d I am speculative\n",tid, *current);
  fflush(stdout);
//  speccode: end debug
#endif
                // Set the current iteration as DONE
                wheel[modvector[*current]] = DONE;

#pragma omp flush(wheel, wheel_ms)

                // If the block being executed by the most speculative
                // thread has an upper iteration that is greater than or
                // equal to the last iteration to perform, then we finish

                if (varblock[1][wheel_ms] >= upper_limit - 1)
                    auxdone = 'T';
                else {
                    // Check if the cell pointed to by wheel_ms+1 is free
#pragma omp flush(wheel, wheel_ms)

                    if (wheel[modvector[wheel_ms + 1]] == FREE) {
                        // Clean the access vector of the cell to be used
                        wheel_data[modvector[wheel_ms + 1]].lastData = 0;
                        // Modify the status of the window cell to be used
                        wheel[modvector[wheel_ms + 1]] = RUNNING;

#pragma omp flush(wheel, wheel_ms, wheel_data)

                        // Increment the most speculative pointer
#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif
                        wheel_ms = wheel_ms + 1;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif

                        /**  nonSpeculative will be 1 if you are the thread
                         **  non speculative before starting the job. Otherwise
                         **  the value will be 0 **/
#pragma omp flush(wheel_ms, wheel_ns)
#ifdef THREADEND_DEBUG
  printf("TID: %d, Actualizando wheel_ms: %d\n",omp_get_thread_num(),wheel_ms);
  fflush(stdout);
#endif
                        *nonSpeculative = (wheel_ms == wheel_ns);

#pragma omp flush(wheel_ms)

                        // Prepare the chunk for running
                        auxjobtodo = 'T';
                        *current = wheel_ms;

                    } // END  if ( wheel[modvector[wheel_ms+1]] == FREE )
                } // END else ( varblock[1,wheel_ms] >= upper_limit)
            } // END else (current == wheel_ns)
        }// END  if ( wheel[modvector[current]] != SQUASHED )
        else {
            //    Squashed thread
#ifdef THREADEND_DEBUG
//  speccode: debug
  printf("threadend: tid= %d, current= %d I was squashed :-(\n",omp_get_thread_num(), *current);
  fflush(stdout);
//  speccode: end debug
#endif
            //    Mark the current chunk as FREE
            wheel[modvector[*current]] = FREE;
            wheel_data[modvector[*current]].lastData = 0;

            auxdone = 'F';
            auxjobtodo = 'F';

#pragma omp flush(wheel)

            // (Note that the ms pointer was moved backwards by the thread that
            // performs the squash operation.)
        } // END else ( wheel[modvector[*current]] != SQUASHED )

    } // END  #pragma OMP CRITICAL

    //***********************************************************************

    // Now check what to do next
    if (auxdone == 'T') {
        flag = JOBDONE;
        return flag;
    }

    if (auxjobtodo == 'T') {
        flag = JOBTODO;
        return flag;
    }

    // If we are here, we cannot start new thread, so spin-wait
#ifdef THREADEND_DEBUG
//  speccode: debug
  printf("threadend: tid= %d, current= %d entering spin wait\n",tid, *current);
  printf("auxdone= %c, auxjobtodo= %c\n", auxdone,auxjobtodo);
  fflush(stdout);
//  speccode: end debug
#endif
#pragma omp flush(wheel, wheel_ms)

e456:
    if (wheel[modvector[wheel_ms + 1]] != FREE) //SPIN1
    {
#pragma omp flush(wheel, wheel_ms)  //SPIN2
#ifdef THREADEND_DEBUG
 printf("threadend: tid= %d, current= %d \n",omp_get_thread_num(), *current);
 fflush(stdout);
#endif
        goto e456; // SPIN3
    }
#ifdef THREADEND_DEBUG
//  speccode: debug
  printf("threadend: tid= %d, current= %d leaving spin wait\n",tid, *current);
  printf("auxdone= %c, auxjobtodo= %c\n", auxdone,auxjobtodo);
  fflush(stdout);
//  speccode: end debug
#endif

    //***********************************************************************
#pragma omp critical
{
        //  Statistics
        SC_threadend2++;

        // Window moved, can start new thread now
        auxdone = 'F';

        // Double check that the new cell is really free and
        // that there are more iterations to run
#pragma omp flush(wheel, wheel_ms)
        if ( (wheel[modvector[wheel_ms + 1]] == FREE) && (varblock[1][wheel_ms] < upper_limit - 1)) {
            // Clean the access vector of the cell to be used
            wheel_data[modvector[wheel_ms + 1]].lastData = 0;

            // Increment the most speculative pointer
#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif
            wheel_ms = wheel_ms + 1;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif

            /**  nonSpeculative will be 1 if you are the thread
             **  non speculative before starting the job. Otherwise
             **  the value will be 0 **/
            #pragma omp flush(wheel_ms, wheel_ns, wheel_data)
#ifdef THREADEND_DEBUG
  printf("TID: %d, Actualizando wheel_ms: %d\n",omp_get_thread_num(),wheel_ms);
  fflush(stdout);
#endif
            *nonSpeculative = (wheel_ms == wheel_ns);

#pragma omp flush(wheel_ms)

            // Prepare the chunk for running
            auxjobtodo = 'T';
            *current = wheel_ms;

            wheel[modvector[wheel_ms]] = RUNNING;

#pragma omp flush(wheel)
        }// END if
        else {
            // The new cell is not actually free, check out if done
#ifdef THREADEND_DEBUG
            //  speccode: debug
            printf("threadend: tid= %d, new cell not free...\n", tid);
            printf("auxdone= %c, auxjobtodo= %c", auxdone, auxjobtodo);
            fflush(stdout);
            //  speccode: end debug
#endif

            // If the block being executed by the most speculative
            // thread has an upper iteration that is greater than or
            // equal to the last iteration to perform, then we finish
#pragma omp flush(wheel_ms)
            if (varblock[1][wheel_ms] >= upper_limit - 1) {
                auxdone = 'T';
                auxjobtodo = 'T';
            } else {
                auxjobtodo = 'F';
                auxdone = 'F';
            }
        }
    } // END #pragma OMP CRITICAL

    //***********************************************************************

    // Spin wait if no job to run

    if (auxjobtodo == 'F') {
#pragma omp flush(wheel, wheel_ms)
#ifdef THREADEND_DEBUG
        //  speccode: debug
        printf("threadend: tid= %d, current= %d re-entering spin wait\n", tid, *current);
        printf("auxdone= %c, auxjobtodo= %c\n", auxdone, auxjobtodo);
        fflush(stdout);
        //  speccode: end debug
#endif
        goto e456;
    }

    // Jump to end if iterations finished

    if (auxdone == 'T') {
        flag = JOBDONE;
        return flag;
    }

    // GO to work
    flag = JOBTODO;
    return flag;

} // END threadend_pointer()

//***********************************************************************
//  commit(): commit operation for aggressive sliding window, without
//            checking for violations and with reduction support
//***********************************************************************

/**
 *
 * @return - the highest position of threads commited
 */
int commit() {
#ifdef COMMIT_DEBUG
    //  speccode: debug
    printf("commit: %d tid=%d I was called\n", current, omp_get_thread_num());
    fflush(stdout);
    //  speccode: end debug
#endif

    // Local variables
    int localtid, mytid;
    int specaux1, specaux2;
    unsigned char *apuntador = NULL;
    unsigned char *apuntador2 = NULL;
    int i = 0;
    threadCopy_structure *punteroMiEstructura = NULL;
    threadCopy_structure miEstructura;

    //  Statistics
    SC_commit++;

    // Mark the iteration as DONE
    wheel[modvector[wheel_ns]] = DONE;
#pragma omp flush(wheel, wheel_ns, wheel_ms)

    // Scan wheel to find most advanced DONE iteration
    for (mytid = wheel_ns; mytid <= (wheel_ms - 1); mytid++) {
        if ((wheel[modvector[mytid]] == DONE) && (wheel[modvector[mytid + 1]] != DONE)) {
            break;
        }
    }

    // Check for corner case: the most speculative can also be DONE
#pragma omp flush(wheel_ms, wheel)
    if ((mytid == (wheel_ms - 1)) && (wheel[modvector[wheel_ms]] == DONE))
        mytid = mytid + 1;

#ifdef COMMIT_DEBUG
    //  speccode: debug
    printf("commit: %d Committing iterations %d to %d\n", current, wheel_ns, mytid);
    fflush(stdout);
    //  speccode: end debug
#endif
#pragma omp flush(wheel_ns)
    for (specaux1 = wheel_ns; specaux1 <= mytid; specaux1++) {

        localtid = modvector[specaux1];
        //#pragma omp flush(wheel_data)
        punteroMiEstructura = wheel_data[localtid].pointer;
        miEstructura = *punteroMiEstructura;
        //#pragma omp flush(wheel_data)
        for (specaux2 = 0; specaux2 < wheel_data[localtid].lastData; specaux2++) {
            // All the elements are distinct to NOTACC

            if ((miEstructura.matriz[T_STATE][specaux2] == ELUP) ||
                    (miEstructura.matriz[T_STATE][specaux2] == UPDATE)) {
                // Commit the updated version
#ifdef COMMIT_DEBUG
                //  speccode: debug
                printf("commit: %d Committing element %d of thread %d\n", current, specaux2, specaux1);
                fflush(stdout);
                //  speccode: end debug
#endif

                apuntador = (unsigned char *) miEstructura.matriz[T_DATA][specaux2];
                apuntador2 = (unsigned char *) miEstructura.matriz[T_POINTER][specaux2];

                // Check data size
                switch (miEstructura.matriz[T_SIZE][specaux2]) {
                    case 1:
                        apuntador2[0] = apuntador[0];
                        break;
                    case 2:
                        apuntador2[0] = apuntador[0];
                        apuntador2[1] = apuntador[1];
                        break;
                    case 4:
                        apuntador2[0] = apuntador[0];
                        apuntador2[1] = apuntador[1];
                        apuntador2[2] = apuntador[2];
                        apuntador2[3] = apuntador[3];
                        break;
                    case 8:
                        apuntador2[0] = apuntador[0];
                        apuntador2[1] = apuntador[1];
                        apuntador2[2] = apuntador[2];
                        apuntador2[3] = apuntador[3];
                        apuntador2[4] = apuntador[4];
                        apuntador2[5] = apuntador[5];
                        apuntador2[6] = apuntador[6];
                        apuntador2[7] = apuntador[7];
                        break;
                    default:
                        for (i = 0; i < miEstructura.matriz[T_SIZE][specaux2]; i++) {
                            apuntador2[i] = apuntador[i];
                        }
                        break;
                }

            }
        } // END for ( specaux2= 0 )

        //    Mark the freed cell as FREE
        wheel[localtid] = FREE;
        wheel_data[localtid].lastData = 0;

#pragma omp flush(wheel, wheel_data)

        // Done with this thread, proceed with next.
    } // END for ( specaux1 = wheel_ns )

    // Free pointers
    punteroMiEstructura = NULL;
    apuntador = NULL;
    apuntador2 = NULL;

    return mytid;

} // END commit()

/***********************************************************************
 **  freeWheel(): releases a slot from wheel
 **  @author Alvaro Estebanez Lopez
 * @param current - Position of the wheel to be reset
 ************************************************************************/
void freeWheel(int current) {
#pragma omp critical
    {
        wheel[modvector[current]] = FREE;
        wheel_data[modvector[current]].lastData = 0;
        //wheel[modvector[current]].maxData = total_pointer;
        //wheel[modvector[current]].pointer = NULL;
#pragma omp flush(wheel, wheel_data)
    }
}

/***********************************************************************
 **  breakEnd(): fires the endLoop flag, squash all successors and commit
                its block
 **  @author Alvaro Estebanez Lopez
 * * @param current - Position of the wheel
 ************************************************************************/
void breakEnd(int current) {
    int squa;

#pragma omp critical
    {
#pragma omp flush(wheel_ms)
        endLoop = 1;
        for (squa = current + 1; squa <= wheel_ms; squa++) {
#pragma omp flush(wheel)
            if (wheel[modvector[squa]] == RUNNING) {
                wheel[modvector[squa]] = SQUASHED;
            }
            if (wheel[modvector[squa]] == DONE) {
                wheel[modvector[squa]] = FREE;
                wheel_data[modvector[squa]].lastData = 0;
            }
#pragma omp flush(wheel, wheel_ms)
        } // END FOR
        // Begin Squash Successors
#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif
        wheel_ms = current;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif
#pragma omp flush(wheel_ms, endLoop)
#ifdef BREAKEND_DEBUG
        printf("TID: %d, Actualizando wheel_ms: %d\n", omp_get_thread_num(), wheel_ms);
        fflush(stdout);
#endif
        // End Squash Successors
        squa = commit();
    }
}

/***********************************************************************
 **  returnEnd(): fires the endReturn flag, squash all successors and commit
                its block
 * * @param current - Position of the wheel to be reset
 **  @author Alvaro Estebanez Lopez
 ************************************************************************/
void returnEnd(int current) {
    int squa, aux;

#pragma omp critical
    {
#pragma omp flush(wheel_ms)
        endReturn = 1;
        for (squa = current + 1; squa <= wheel_ms; squa++) {
#pragma omp flush(wheel)
            if (wheel[modvector[squa]] == RUNNING) {
                wheel[modvector[squa]] = SQUASHED;
            }
            if (wheel[modvector[squa]] == DONE) {
                wheel[modvector[squa]] = FREE;
                wheel_data[modvector[squa]].lastData = 0;
            }
#pragma omp flush(wheel, wheel_ms)
        } // END FOR
        /* Begin Squash Successors */
#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif
        wheel_ms = current;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif
#pragma omp flush(wheel_ms, endReturn)
#ifdef RETURNEND_DEBUG
        printf("TID: %d, Actualizando wheel_ms: %d\n", omp_get_thread_num(), wheel_ms);
        fflush(stdout);
#endif
        /* End Squash Successors */
        aux = commit();
#ifdef WHEEL_NS_DEBUG
    printf("cambio en wheel_ns: ANTES=%d,",wheel_ns);
    fflush(stdout);
#endif
        wheel_ns = aux + 1;
#ifdef WHEEL_NS_DEBUG
    printf(" DESPUES=%d \n",wheel_ns);
    fflush(stdout);
#endif
#pragma omp flush(wheel_ns)
#ifdef RETURNEND_DEBUG
        printf("RETURN END: TID: %d, Actualizando wheel_ns: %d\n", omp_get_thread_num(), wheel_ns);
        fflush(stdout);
#endif
    }
}

/***********************************************************************
 **  squashThreads(): squash all threads from current to wheel_ms
 **  @author Alvaro Estebanez Lopez
 * * @param current - Position of the wheel to be reset
 ************************************************************************/
void squashThreads(int current) {
    int squa;

#pragma omp critical
    {
        if ((wheel[modvector[current]] == RUNNING) && (current <= wheel_ms)) {
            // Squash "specaux1" and all its successors.
            // If a sucessor is RUNNING, this is set as SQUASHED.
            // If a sucessor is DONE, then is set as FREE.

#pragma omp flush(wheel_ms)
            for (squa = current; squa <= wheel_ms; squa++) {
#ifdef SQUASHTHREADS_DEBUG
                printf("squashThreads %d \n", squa);
                fflush(stdout);
#endif
#pragma omp flush(wheel)
                if (wheel[modvector[squa]] == RUNNING) {
                    wheel[modvector[squa]] = SQUASHED;
                }
                if (wheel[modvector[squa]] == DONE) {
                    wheel[modvector[squa]] = FREE;
                    wheel_data[modvector[squa]].lastData = 0;
                }

#pragma omp flush(wheel, wheel_ms)
            } // END FOR
#ifdef WHEEL_MS_DEBUG
    printf("\tcambio en wheel_ms: ANTES=%d,",wheel_ms);
    fflush(stdout);
#endif
            wheel_ms = current - 1;
#ifdef WHEEL_MS_DEBUG
    printf(" DESPUES=%d \n",wheel_ms);
    fflush(stdout);
#endif
#pragma omp flush(wheel_ms)
#ifdef SQUASHTHREADS_DEBUG
            printf("TID: %d, Actualizando wheel_ms: %d\n", omp_get_thread_num(), wheel_ms);
            fflush(stdout);
#endif
        }

    } // END OMP CRITICAL
}

//***********************************************************************
//   initializeCounters(): operation to initialize counters
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************

void initializeCounters() {
    spec_violations = 0;
    spec_squashes = 0;
#ifdef exceptions
    countExceptions = 0;
    exceptionsBlocksRestart = 0;
#endif
    SC_commit = 0;
    SC_threadend1 = 0;
    SC_threadend2 = 0;
    SC_specstore = 0;
    //access_specload = 0;
}

//***********************************************************************
//  specepilogue(): Show total number of violations and squashes.
//            (should be called as last instruction of the application)
//***********************************************************************

void specepilogue() {
    //  Main header file with all common variables
    //    #include appcommon.h
    //    #include <stdio.h>

    int i;

    printf("------------------------------------------\n");
    printf("Block number\tblock size\tnumber of executions\n");

    for (i = 0; i <= maxblock + 10; i++)
        printf("\t%d\t\t%d\t\t%d\n", i, varblock[1][i] - varblock[0][i] + 1, varblock[2][i]);

    printf("------------------------------------------\n");

    printf("Total violations: %d ; Total squashes: %d\n", spec_violations, spec_squashes);
    printf("Access specstore SC: %d \n", SC_specstore);
    printf("Access threadend1 SC: %d \n", SC_threadend1);
    printf("Access threadend2 SC: %d \n", SC_threadend2);
    printf("Access commit SC: %d \n", SC_commit);
} // END specepilogue()
