/**
 * @author Alvaro Estebanez Lopez
 * This macro should be used inside a speculative loop
 * instead of continue();
 **/

#define continueSpecEngine(IDLOOP) 								\
	goto labelEndIteration_##IDLOOP;								

/** 
 * @author Alvaro Estebanez Lopez
 * Finishing the loop because of a break instruction only if it is executed by
 * a no speculative thread. Otherwise, the thread is squashed with its successors
 **/
#define breakSpecEngine(IDLOOP)								\
        if (nonSpeculative == 1)                                                        \
	{										\
	  breakEnd(current);                                                            \
	  /*setEndLoop();								\
	  squashSuccessors(current+1);							\
	  threadend( &current,(int *) vector, &my_IV_tail);				*/\
	  goto labelEndLoop_##IDLOOP; 									\
	}										\
	else										\
	{										\
	  squashThreads(current);                                                       \
	}	

/** 
 * @author Alvaro Estebanez Lopez
 * @arg state returned value by the exit instruction
 * Finishing the program because of a exit instruction only if it is executed by
 * a no speculative thread. Otherwise, the thread is squashed with its successors
 **/
#define exitSpecEngine(state)								\
        if (nonSpeculative == 1)                                                        \
	{										\
	  breakEnd(current);					\
	  /*sumatorio = 0;			\
  	   Just to check the example results
          for( i = 0; i < MAX; i++)		\
          {					\
           sumatorio = sumatorio + vector[i];	\
          }					\
          printf("%d\n",sumatorio);		*/\
	  exit(state);									\
	}										\
	else										\
	{										\
	  squashThreads(current);                                                       \
	}	

/** 
 * @author Alvaro Estebanez Lopez
 * Finishing the loop because of a return instruction only if it is executed by
 * a no speculative thread. Otherwise, the thread is squashed with its successors
 * @arg IDLOOP loop identifier
 **/
#define returnSpecEngine(IDLOOP)								\
        /*if (isNonSpeculative()) */                                                        \
        if (nonSpeculative == 1)                                                         \
	{										\
	  returnEnd(current);					\
	  /*setEndLoop();								\
	  squashSuccessors(current+1);							\
	  threadend( &current,(int *) vector, &my_IV_tail);				*/\
	  goto labelEndLoop_##IDLOOP; 									\
	}										\
	else										\
	{										\
	  squashThreads(current);                                                       \
	}	

/** 
 * @author Alvaro Estebanez Lopez
 * This macro must be called below the end of a speculative loop inside a function.
 * This macro enables to know if a return
 * instruction has been invoked by the no speculative thread. As OpenMP do not permit to
 * invoke a return instruction inside a loop, in this way the return is invoked outside
 * the loop.
 **/
#define testReturnSpecEngine()								\
        if (endReturn)		                                                        \
	  return;

/** 
 * @author Alvaro Estebanez Lopez
 * @arg iteration index of the iteration is being excuted inside the speculative loop
 * @arg ini initial index of the loop 
 * @arg IDLOOP loop identifier
 * Assigning the first block each thread
 * According to flag exceptions checkPoint is enabled or disabled
 **/

//#define initLoopSpecEngine(iteration,ini,IDLOOP)			\
    for ( tid = 0 ; tid <= threads-1 ; tid++ )						\
    {											\
      /*printf("INI %d\n",omp_get_thread_num());*/				 	\
      /*if ( tid == 0 )									\
        nonSpeculative = 1;								\
      current = tid;*/									\
      if ( getFirstBlock(&current, &nonSpeculative) == 0)                               \
      {											\
        /*printf("NO TIENES TRABAJO %d \n",omp_get_thread_num());*/			\
	goto labelEndLoop_##IDLOOP;							\
      }											\
      iteration = varblock[0][current]+ini;						\
      varblock[2][current] = varblock[2][current]+1;					\
printf("\t\tvarblock[0][current]:%d,current:%d,iteration:%d\n",varblock[0][current],current,iteration);\
      /*if ( iteration >= nIterations)							\
	goto labelEndLoop_##IDLOOP; 							*/\
labelStartIteration_##IDLOOP:		printf("ENDLOOP,iteration:%d,current:%d\n",iteration,current);


/** 
 * @author Alvaro Estebanez Lopez
 * @arg iteration index of the iteration is being excuted inside the speculative loop
 * @arg nIterations total number of iterations in the loop
 * @arg ini initial index of the loop 
 * @arg IDLOOP loop identifier
 * Returning point after ending an iteration
 * A new block of iterations will be assigned providing that
 * there are still work to execute
 **/
//#define endLoopSpecEngine(iteration, nIterations, ini, IDLOOP)  \
	/**********************************************************************		\
	// speccode: end of thread and commit						\
	//           First check if block of iterations is done **************/		\
	if (endLoop || endReturn) 							\
        {										\
	  freeWheel(current);								\
          goto labelEndLoop_##IDLOOP;							\
        }										\
        labelEndIteration_##IDLOOP:							\
     	if ( (iteration != varblock[1][current]+ini) && (iteration < nIterations-1) )	\
	{			 							\
		iteration = iteration + 1;\
		goto labelStartIteration_##IDLOOP;					\
	}										\
	/* Thread done, perform the commit */						\
	labelSquash_##IDLOOP:								\
                retflag = threadend_pointer(&current, &nonSpeculative);		\
	/* End if job done */								\
	if ( retflag == JOBDONE ) goto labelEndLoop_##IDLOOP;				\
	if ( retflag == JOBTODO )							\
	{										\
	/*  Set loop variable */							\
		iteration = varblock[0][current] + ini ;				\
		varblock[2][current] = varblock[2][current] + 1;			\
                goto labelStartIteration_##IDLOOP;					\
	}										\
	labelEndLoop_##IDLOOP: ; /*freeWheel(current);*/					\
	} /*End for initLoop()*/							\
    //} //End pragma omp


/** 
 * @author Alvaro Estebanez Lopez
 * @arg IDLOOP loop identifier
 * Going to the end of the loop to call threadend function
 **/
#define earlySquash(IDLOOP)  \
        {\
        /*printf("SQUASH\n");*/\
	goto labelSquash;\
	}
