/*
 * incr-2d: C version of 2-dimensional, incremental convex hull
 * Alvaro Estebanez Lopez, Diego R. Llanos
 * University of Valladolid
 * Started in Wed May 23 16:15:00 CET 2012
 */

#include <stdio.h>
#include <stdlib.h>

//***********************************************************************
// Speccode: include files for timing
// #include "libcollector.h"
#include <sys/time.h>
//***********************************************************************

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file 
#include "specEngine.h"
//***********************************************************************

/*--------------------------------------------------------------------------*/
// speccode: handler
//newException(int_handler);
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
// speccode: to measure time in UNIX environments
void gettime(double *);
double spec_time = 0;
/*--------------------------------------------------------------------------*/


// N total number of points  (min. 3)
#define N 10000000

#define source 1
#define target 2
#define succ   3
#define pred   4
#define inch   5

#define x 1
#define y 2

#define TOP 4400

int hull[TOP + 1][5 + 1];
double input[2 + 1][N + 1];

int i, j, k;
int p, psource, ptarget;
int auxinch;
int lastedge;
int e, eprev, enext, ea;
int first;
double px, py, qx, qy, dx, dy;
double determ;
int outside;

// Function
double rightturn(double dx, double dy, double px, double py, double qx, double qy);

int main() {
    FILE *inputfile;

    /************************************************************************/
    // speccode: Stop profiling
    // collector_pause();
    /************************************************************************/
    //***********************************************************************
    // speccode: Initializing structures
    specinit();
    //***********************************************************************


    //  Opens the input file
    //if (inputfile = fopen("/home/alvaro/Documents/spec/Points/input/disc.dat", "r"))
    if (inputfile = fopen("disc.in", "r")) {
        int not_used;
        for (i = 1; i < N + 1; i++) {

            not_used = fscanf(inputfile, "%lf %lf", &input[x][i], &input[y][i]);
            //printf("New point: %lf %lf\n",input[x][i],input[y][i]);
        }

        fclose(inputfile);
    } else {
        printf("Error reading from file \n ");
        exit(0);
    }

    // Calculate the first determinant
    dx = input[x][1];
    dy = input[y][1];
    px = input[x][2];
    py = input[y][2];
    qx = input[x][3];
    qy = input[y][3];

    determ = rightturn(dx, dy, px, py, qx, qy);

    if (determ == 0) {
        printf("incremental: The first three points are aligned.");
        printf("incremental: We aree too lazy to proceed. Stop.");
        exit(0);
    }

    // If determinant < 0, reorder points two and three
    if (determ < 0) {
        input[x][2] = qx;
        input[y][2] = qy;
        input[x][3] = px;
        input[y][3] = py;
    }

    // Links the first edge
    hull[1][source] = 1;
    hull[1][target] = 2;
    hull[1][pred] = 3;
    hull[1][succ] = 2;
    hull[1][inch] = 1;

    // Links the second edge
    hull[2][source] = 2;
    hull[2][target] = 3;
    hull[2][pred] = 1;
    hull[2][succ] = 3;
    hull[2][inch] = 1;

    // Links the third edge
    hull[3][source] = 3;
    hull[3][target] = 1;
    hull[3][pred] = 2;
    hull[3][succ] = 1;
    hull[3][inch] = 1;

    hull[TOP][1] = 1;
    hull[TOP][5] = 3;
    // Put -1 in all remaining elements of hull, just in case.
    for (i = 4; i < TOP; i++) {
        for (j = 1; j < 6; j++) {
            //-1 can prompt a SEGV signal
            //hull[i][j] = -1;
            hull[i][j] = 0;
        }
    }

    // Main loop (iterating over remaining points)

    //***********************************************************************
    // speccode: OMP threading directive
    omp_set_num_threads(threads);
    // speccode: initialize speculation structrures
    specstart(N + 1 - 3);
    //***********************************************************************

    /************************************************************************/
    // speccode: Start profiling
    double spec_start, spec_stop;
    gettime(&spec_start);
    /************************************************************************/

    //***********************************************************************
    // speccode: Start profiling
    // collector_resume();
    //***********************************************************************
    //***********************************************************************
    // speccode: Speculative loop
    int iteration, ini, nIterations;
    int value;

#pragma omp parallel default(none)                                  \
   private(outside,p,px,py,qx,qy,dx,dy,determ,i,k,lastedge, \
      ini,nIterations,iteration,psource,ptarget,auxinch,e,eprev,enext,ea,value,  \
      current, tid, retflag,nonSpeculative,stdout)               \
   shared(input,hull, \
      endLoop, endReturn, wheel_ns, wheel_ms,    \
      wheel,           \
      upper_limit, varblock)                            
    {
        //setException(int_handler,SIGFPE);
#pragma omp for \
   schedule(static)
        //for ( i = 4 ; i < N+1 ; i++)
        // Format: initLoopSpecEngine(vector, iterator, ini,IDLOOP)
        //     initLoopSpecEngine(hull,k,1,1);
        for (tid = 0; tid <= threads - 1; tid++) {

            ini = 1;

            //printf("incr-2d: Calling getFirstBlock().\n");
            //fflush(stdout);

            if (getFirstBlock(&current, &nonSpeculative) == 0) {
                /*printf("NO TIENES TRABAJO %d \n",omp_get_thread_num());*/
                goto labelEndLoop;
            }

            //printf("incr-2d: getFirstBlock() called finished.\n");
            //fflush(stdout);

            iteration = varblock[0][current] + ini;
            varblock[2][current] = varblock[2][current] + 1;
            //printf("\t\tvarblock[0][current]:%d,current:%d,iteration:%d\n",varblock[0][current],current,iteration);
labelStartIteration: //printf("ENDLOOP,iteration:%d,current:%d\n",iteration,current);

            k = iteration;
            // current = tid;
            // my_IV_tail = 0;
            // k = varblock[0][current]+1;
            // varblock[2][current] = varblock[2][current]+1;

            i = k + 3;
            // Debug
            // if ((i/100000)*100000==i)
            // printf(".");
            // End Debug

            // Debug
            // printf("incr-2d: starting iteration %i, attempt %i\n",i, varblock[2][current]);
            // End Debug

            outside = 0;
            for (e = 1; e < 4; e++) {
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      psource = hull[e][source];
                if (specload_pointer((unsigned char *) &(hull[e][source]), sizeof (hull[e][source]), current
                        , (unsigned char *) &value) == -1)
                    earlySquash(1);
                psource = value;
                //***********************************************************************
                dx = input[x][psource];
                dy = input[y][psource];


                //***********************************************************************
                // speccode: speculative load. Original line:
                //      ptarget = hull[e][target];
                if (specload_pointer((unsigned char *) &(hull[e][target]), sizeof (hull[e][target]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                ptarget = value;
                //***********************************************************************
                px = input[x][ptarget];
                py = input[y][ptarget];

                qx = input[x][i];
                qy = input[y][i];

                determ = rightturn(dx, dy, px, py, qx, qy);
                if (determ < 0) {
                    // The point is outside
                    outside = 1;
                    break;
                }
            }

e500:
            //***********************************************************************
            // speccode: speculative load. Original line:
            //      auxinch = hull[e][inch];
            if (specload_pointer((unsigned char *) &(hull[e][inch]), sizeof (hull[e][inch]), current, (unsigned char *) &value) == -1)
                earlySquash(1);
            auxinch = value;
            //***********************************************************************
            if ((auxinch == 0) && (outside)) {
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      eprev = hull[e][pred];
                if (specload_pointer((unsigned char *) &(hull[e][pred]), sizeof (hull[e][pred]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                eprev = value;
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      psource = hull[eprev][source];
                if (specload_pointer((unsigned char *) &(hull[eprev][source]), sizeof (hull[eprev][source]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                psource = value;
                //***********************************************************************

                dx = input[x][psource];
                dy = input[y][psource];

                //***********************************************************************
                // speccode: speculative load. Original line:
                //      ptarget = hull[eprev][target];
                if (specload_pointer((unsigned char *) &(hull[eprev][target]), sizeof (hull[eprev][target]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                ptarget = value;
                //***********************************************************************
                px = input[x][ptarget];
                py = input[y][ptarget];

                qx = input[x][i];
                qy = input[y][i];

                determ = rightturn(dx, dy, px, py, qx, qy);

                if (determ < 0)
                    e = eprev;
                else {
                    //***********************************************************************
                    // speccode: speculative load. Original line:
                    //        enext = hull[e][succ];
                    if (specload_pointer((unsigned char *) &(hull[e][succ]), sizeof (hull[e][succ]), current, (unsigned char *) &value) == -1)
                        earlySquash(1);
                    enext = value;
                    //***********************************************************************
                    //***********************************************************************
                    // speccode: speculative load. Original line:
                    //        psource = hull[enext][source];
                    if (specload_pointer((unsigned char *) &(hull[enext][source]), sizeof (hull[enext][source]), current, (unsigned char *) &value) == -1)
                        earlySquash(1);
                    psource = value;
                    //***********************************************************************

                    dx = input[x][psource];
                    dy = input[y][psource];

                    //***********************************************************************
                    // speccode: speculative load. Original line:
                    //        ptarget = hull[enext][target];
                    if (specload_pointer((unsigned char *) &(hull[enext][target]), sizeof (hull[enext][target]), current, (unsigned char *) &value) == -1)
                        earlySquash(1);
                    ptarget = value;
                    //***********************************************************************
                    px = input[x][ptarget];
                    py = input[y][ptarget];

                    qx = input[x][i];
                    qy = input[y][i];

                    determ = rightturn(dx, dy, px, py, qx, qy);

                    if (determ >= 0)
                        outside = 0;
                    else
                        e = enext;
                }
                goto e500;
            }

            if (outside) {
                //Compute upper tangent
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      enext = hull[e][succ];
                if (specload_pointer((unsigned char *) &(hull[e][succ]), sizeof (hull[e][succ]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                enext = value;
                //***********************************************************************

e600:
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      psource = hull[enext][source];
                if (specload_pointer((unsigned char *) &(hull[enext][source]), sizeof (hull[enext][source]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                psource = value;
                //***********************************************************************
                dx = input[x][psource];
                dy = input[y][psource];

                //***********************************************************************
                // speccode: speculative load. Original line:
                //      ptarget = hull[enext][target];
                if (specload_pointer((unsigned char *) &(hull[enext][target]), sizeof (hull[enext][target]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                ptarget = value;
                //***********************************************************************
                px = input[x][ptarget];
                py = input[y][ptarget];

                qx = input[x][i];
                qy = input[y][i];

                determ = rightturn(dx, dy, px, py, qx, qy);

                //***********************************************************************
                // speccode: speculative load. Original line:
                //      lastedge = hull[TOP][5];
                if (specload_pointer((unsigned char *) &(hull[TOP][5]), sizeof (hull[TOP][5]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                lastedge = value;
                //***********************************************************************

                if (determ <= 0) {
                    //***********************************************************************
                    // speccode: speculative store. Original line:
                    //        hull[enext][inch] = 0;
                    value = 0;
                    specstore_pointer((unsigned char *) &(hull[enext][inch]), sizeof (hull[enext][inch]), current, (unsigned char *) &value);
                    //***********************************************************************
                    //***********************************************************************
                    // speccode: speculative store. Original line:
                    //        hull[enext][pred] = lastedge+1;
                    value = lastedge + 1;
                    specstore_pointer((unsigned char *) &(hull[enext][pred]), sizeof (hull[enext][pred]), current, (unsigned char *) &value);
                    //***********************************************************************
                    //***********************************************************************
                    // speccode: speculative load. Original line:
                    //        ea = hull[enext][succ];
                    if (specload_pointer((unsigned char *) &(hull[enext][succ]), sizeof (hull[enext][succ]), current, (unsigned char *) &value) == -1)
                        earlySquash(1);
                    ea = value;
                    //***********************************************************************
                    //***********************************************************************
                    // speccode: speculative store. Original line:
                    //        hull[enext][succ] = lastedge+2;
                    value = lastedge + 2;
                    specstore_pointer((unsigned char *) &(hull[enext][succ]), sizeof (hull[enext][succ]), current, (unsigned char *) &value);
                    //***********************************************************************

                    enext = ea;
                    goto e600;
                }
                // Compute lower tangent

                //***********************************************************************
                // speccode: speculative load. Original line:
                //      eprev = hull[e][pred];
                if (specload_pointer((unsigned char *) &(hull[e][pred]), sizeof (hull[e][pred]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                eprev = value;
                //***********************************************************************
e700:
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      psource = hull[eprev][source];
                if (specload_pointer((unsigned char *) &(hull[eprev][source]), sizeof (hull[eprev][source]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                psource = value;
                //***********************************************************************
                dx = input[x][psource];
                dy = input[y][psource];

                //***********************************************************************
                // speccode: speculative load. Original line:
                //      ptarget = hull[eprev][target];
                if (specload_pointer((unsigned char *) &(hull[eprev][target]), sizeof (hull[eprev][target]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                ptarget = value;
                //***********************************************************************
                px = input[x][ptarget];
                py = input[y][ptarget];

                qx = input[x][i];
                qy = input[y][i];

                determ = rightturn(dx, dy, px, py, qx, qy);

                if (determ <= 0) {
                    //***********************************************************************
                    // speccode: speculative store. Original line:
                    //        hull[eprev][inch] = 0;
                    value = 0;
                    specstore_pointer((unsigned char *) &(hull[eprev][inch]), sizeof (hull[eprev][inch]), current, (unsigned char *) &value);
                    //***********************************************************************
                    //***********************************************************************
                    // speccode: speculative load. Original line:
                    //        ea = hull[eprev][pred];
                    if (specload_pointer((unsigned char *) &(hull[eprev][pred]), sizeof (hull[eprev][pred]), current, (unsigned char *) &value) == -1)
                        earlySquash(1);
                    ea = value;
                    //***********************************************************************
                    //***********************************************************************
                    // speccode: speculative store. Original line:
                    //        hull[eprev][pred] = lastedge+1;
                    value = lastedge + 1;
                    specstore_pointer((unsigned char *) &(hull[eprev][pred]), sizeof (hull[eprev][pred]), current, (unsigned char *) &value);
                    //***********************************************************************
                    //***********************************************************************
                    // speccode: speculative store. Original line:
                    //        hull[eprev][succ] = lastedge+2;
                    value = lastedge + 2;
                    specstore_pointer((unsigned char *) &(hull[eprev][succ]), sizeof (hull[eprev][succ]), current, (unsigned char *) &value);
                    //***********************************************************************
                    eprev = ea;
                    goto e700;
                }
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[e][inch] = 0;
                value = 0;
                specstore_pointer((unsigned char *) &(hull[e][inch]), sizeof (hull[e][inch]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[e][pred] = lastedge+1;
                value = lastedge + 1;
                specstore_pointer((unsigned char *) &(hull[e][pred]), sizeof (hull[e][pred]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[e][succ] = lastedge+2;
                value = lastedge + 2;
                specstore_pointer((unsigned char *) &(hull[e][succ]), sizeof (hull[e][succ]), current, (unsigned char *) &value);
                //***********************************************************************

                // Add new edges to convex hull

                lastedge = lastedge + 1;
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[TOP][5] = lastedge;
                value = lastedge;
                specstore_pointer((unsigned char *) &(hull[TOP][5]), sizeof (hull[TOP][5]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      hull[lastedge][source] = hull[eprev][target];
                if (specload_pointer((unsigned char *) &(hull[eprev][target]), sizeof (hull[eprev][target]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                // speccode: speculative store. Original line:
                specstore_pointer((unsigned char *) &(hull[lastedge][source]), sizeof (hull[lastedge][source]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][target] = i;
                value = i;
                specstore_pointer((unsigned char *) &(hull[lastedge][target]), sizeof (hull[lastedge][target]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][succ] = lastedge + 1;
                value = lastedge + 1;
                specstore_pointer((unsigned char *) &(hull[lastedge][succ]), sizeof (hull[lastedge][succ]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][pred] = eprev;
                value = eprev;
                specstore_pointer((unsigned char *) &(hull[lastedge][pred]), sizeof (hull[lastedge][pred]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][inch] = 1;
                value = 1;
                specstore_pointer((unsigned char *) &(hull[lastedge][inch]), sizeof (hull[lastedge][inch]), current, (unsigned char *) &value);
                //***********************************************************************

                lastedge = lastedge + 1;
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[TOP][5] = lastedge;
                value = lastedge;
                specstore_pointer((unsigned char *) &(hull[TOP][5]), sizeof (hull[TOP][5]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][source] = i;
                value = i;
                specstore_pointer((unsigned char *) &(hull[lastedge][source]), sizeof (hull[lastedge][source]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative load. Original line:
                //      hull[lastedge][target] = hull[enext][source];
                if (specload_pointer((unsigned char *) &(hull[enext][source]), sizeof (hull[enext][source]), current, (unsigned char *) &value) == -1)
                    earlySquash(1);
                // speccode: speculative store. Original line:
                specstore_pointer((unsigned char *) &(hull[lastedge][target]), sizeof (hull[lastedge][target]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][succ] = enext;
                value = enext;
                specstore_pointer((unsigned char *) &(hull[lastedge][succ]), sizeof (hull[lastedge][succ]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][pred] = lastedge - 1;
                value = lastedge - 1;
                specstore_pointer((unsigned char *) &(hull[lastedge][pred]), sizeof (hull[lastedge][pred]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[lastedge][inch] = 1;
                value = 1;
                specstore_pointer((unsigned char *) &(hull[lastedge][inch]), sizeof (hull[lastedge][inch]), current, (unsigned char *) &value);
                //***********************************************************************

                // Connect to old convex hull
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[enext][pred] = lastedge;
                value = lastedge;
                specstore_pointer((unsigned char *) &(hull[enext][pred]), sizeof (hull[enext][pred]), current, (unsigned char *) &value);
                //***********************************************************************
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[eprev][succ] = lastedge - 1;
                value = lastedge - 1;
                specstore_pointer((unsigned char *) &(hull[eprev][succ]), sizeof (hull[eprev][succ]), current, (unsigned char *) &value);
                //***********************************************************************

                //Update first
                //***********************************************************************
                // speccode: speculative store. Original line:
                //      hull[TOP][1] = lastedge;
                value = lastedge;
                specstore_pointer((unsigned char *) &(hull[TOP][1]), sizeof (hull[TOP][1]), current, (unsigned char *) &value);
                //***********************************************************************
            } //outside==0
            // End of process, proceed with next point

            // **********************************************************************
            // speccode: end of thread and commit
            //endLoopSpecEngine( (int *)hull,k,N+1-3,4,1);
            //endLoopSpecEngine(i, cur_upper_limit, 0, 1);
            //     iteration = k;
            nIterations = N + 1 - 3;
            ini = 4;
            /**
             * @author Alvaro Estebanez Lopez
             * @arg iteration index of the iteration is being excuted inside the speculative loop
             * @arg nIterations total number of iterations in the loop
             * @arg ini initial index of the loop
             * @arg IDLOOP loop identifier
             * Returning point after ending an iteration
             * A new block of iterations will be assigned providing that
             * there are still work to execute
             **/
            //#define endLoopSpecEngine(iteration, nIterations, ini, IDLOOP)
            /**********************************************************************		\
            // speccode: end of thread and commit						\
            //           First check if block of iterations is done **************/		\
	if (endLoop || endReturn) {
                freeWheel(current);
                goto labelEndLoop;
            }
labelEndIteration:
            if ((iteration != varblock[1][current] + ini) && (iteration < nIterations - 1)) {
                iteration = iteration + 1;
                goto labelStartIteration;
            }
            /* Thread done, perform the commit */
labelSquash:
            retflag = threadend_pointer(&current, &nonSpeculative);
            /* End if job done */
            if (retflag == JOBDONE) goto labelEndLoop;
            if (retflag == JOBTODO) {
                /*  Set loop variable */
                iteration = varblock[0][current] + ini;
                varblock[2][current] = varblock[2][current] + 1;
                goto labelStartIteration;
            }
labelEndLoop:
            ; /*freeWheel(current);*/
        } /*End for initLoop()*/

    } // END OMP PARALLEL

    /************************************************************************/
    // speccode: Stop measuring time
    gettime(&spec_stop);
    spec_time += spec_stop - spec_start;
    /************************************************************************/

    // Show the results

    // e = hull[TOP][1];
    // p = hull[e][source];
    // e400:
    // printf("%f %f\n",input[x][p],input[y][p]);
    // e = hull[e][succ];
    // p = hull[e][source];
    // if ( e != hull[TOP][1] ) 
    // goto e400;

    printf("Total number of edges: %d\n", hull[TOP][5]);

    // speccode: statistics for target loop.
    specepilogue(2);

    // -----------------------------------------------------------
    // speccode: print time results
    fprintf(stderr, "--------------------------------------------------\n");
    fprintf(stderr, "speccode: Time elapsed: %lf seconds.\n", spec_time);
    // -----------------------------------------------------------

    return 0;

}

double rightturn(double dx, double dy, double px, double py, double qx, double qy) {
    return (dx * py + dy * qx + px * qy - qx * py - dx * qy - px * dy);
}

/*--------------------------------------------------------------------------*/
// speccode: time measurement

void gettime(double *t) {
    struct timeval tv;
    gettimeofday(&tv, (void *) 0);
    *t = tv.tv_sec + 1.0e-6 * tv.tv_usec;
}
