
/* =====================================================================
   mec-2d: C version of 2-dimensional Minimun Enclosing Circle.
   Alvaro Estebanez Lopez
   Original Fortran version: Belen Palop, Diego R. Llanos (members of the GeoPar group)
   University of Valladolid
   Revision: April 2011
   Started in Wed Apr 13 17:35:00 2011
   ===================================================================== */

// =====================================================================

#include <stdio.h>
#include <stdlib.h>
// library use to timing K loop
#include "time.h"

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//***********************************************************************

// =====================================================================
//	constants

//	N: total number of points (min. 1)
#define N 10000000

//      EPS: Absolute precision threshold
#define EPS 0.001

//      centrox, centroy, cuad_radio: indexes of the RESULT structure.
#define centrox 0
#define centroy 1
#define cuad_radio 2
#define solpx 3
#define solpy 4
#define solqx 5
#define solqy 6
#define solrx 7
#define solry 8
#define puntdef 9


// =====================================================================
//	Data structures and variables

//	Loop counters
int i, j, k;

//      result: vector to hold current results
double result[10];

//	Input data matrix
double input[2][N];

//	Old Result variables (now used to temporarily hold result[] values)
double spec_centrox, spec_centroy, spec_cuad_radio;
double spec_solpx, spec_solpy;
double spec_solqx, spec_solqy;
double spec_solrx, spec_solry;
double spec_puntdef;
double value;
int linear;
//	File descriptor
FILE *inputfile;

//	Auxiliar variables
double px, py, qx, qy, rx, ry;
double yqr, ypr, xqr, xpr, xrp, xqp, xpq, ypq;
double lambda1, lambda2, denom;
int pointsRead;
int x, y;

// =====================================================================
//      Functions

double cuad_dist(double px, double py, double qx, double qy);
//newException(int_handler);

// =====================================================================

int main(int argc, char* argv[]) {


    // =====================================================================
    //	Open and read input file
    if (inputfile = fopen("/home/Points/input/disc-mec.in", "r")) {

        for (pointsRead = 0; pointsRead < N; pointsRead++) {
            if (fscanf(inputfile, "%d %d", &x, &y) < 0) {
                printf("Error reading from file\n");
                exit(-1);
            }
            input[0][pointsRead] = x;
            input[1][pointsRead] = y;
        }//for
    } else {
        printf("Error reading from file\n");
        exit(-1);
    }

    fclose(inputfile);


    //***********************************************************************
    // speccode: Stop profiling
    // collector_pause();
    //***********************************************************************
    // speccode: Initializing structures
    specinit();
    initializeCounters();
    //***********************************************************************


// --------------------------------------------------------------------
//	Initializations.
 result[centrox]=0;
 result[centroy]=0;
 result[cuad_radio]=0;

// --------------------------------------------------------------------
//	Beginning of process 
 for (i = 0; i < N; i++){
  px=input[0][i];
  py=input[1][i];

  if (cuad_dist(px,py,result[centrox],result[centroy])-result[cuad_radio] > EPS){

   // Debug 
   //  printf("mec-2d: Nuevo punto de fallo: %d\n", i);
   // End debug

   // Inlined version of circunf1p() follows 
   result[centrox]=px;
   result[centroy]=py;
   result[cuad_radio]=0;
   result[solpx]=px;
   result[solpy]=py;
   result[puntdef]=1;

   // print *,"New upper limit for middle loop: ",(i-1)

   for (j = 0; j < (i-1); j++){
    qx=input[0][j];
    qy=input[1][j];
     if (cuad_dist(qx,qy,result[centrox],result[centroy])-result[cuad_radio] > EPS){

      // Inlined version of circunf2p() follows 
      result[centrox]=(px+qx)/2;
      result[centroy]=(py+qy)/2;
      result[cuad_radio]=cuad_dist(px,py,qx,qy)/4;
      result[solpx]=px;
      result[solpy]=py;
      result[solqx]=qx;
      result[solqy]=qy;
      result[puntdef]=2;
/*
				// Debug: Check if result is correct 
				if ( cuad_dist(qx, qy, result[centrox], result[centroy]) - result[cuad_radio] > EPS){
					printf("Error in second loop!\n");
					printf("Center is ('%f', '%f')\n",result[centrox], result[centroy]);
					printf("Squared radius is %f", result[cuad_radio]);
					exit(-1);
				}
*/
//***********************************************************************
// speccode: OMP threading directive
    omp_set_num_threads(threads);
// speccode: initialize speculation structures
    specstart(j);
// speccode: variable upper limit inicialization
    //upper_limit = j;
//***********************************************************************
//***********************************************************************
// speccode: Start profiling
    // collector_resume();
//***********************************************************************
// start time counter
startTime();

    int iteration, ini, nIterations;
    double DOUBLE_VALUE = 0;

    // start time counter
    startTime();
    //***********************************************************************
    // speccode: Speculative loop
#pragma omp parallel default(none)                                  \
               private(rx,ry,k,iteration, \
 		       yqr, ypr, xqr, xpr, xrp, xqp,xpq, ypq, inicio, fin, \
		       lambda1, lambda2,denom, spec_centrox, spec_centroy, \
		       spec_cuad_radio, spec_solpx,spec_solpy,spec_solqx, \
		       spec_solqy,spec_solrx,spec_solry,spec_puntdef, \
                       value,linear,current, tid, retflag, \
		       stdout,ini,nIterations, DOUBLE_VALUE, nonSpeculative)    \
              shared(px,py,qx,qy,i,j,input,result, wheel_ns, wheel_ms, \
                     wheel,wheel_data, counter, \
                     upper_limit, varblock, endLoop, endReturn)
    {

        //setException(int_handler, SIGSEGV);

#pragma omp for \
   schedule(static)

     //                   for (k = 0; k < j ; k++) {
/**
         * @author Alvaro Estebanez Lopez
         * @arg iteration index of the iteration is being excuted inside the speculative loop
         * @arg ini initial index of the loop
         * @arg IDLOOP loop identifier
         * Assigning the first block each thread
         * According to flag exceptions checkPoint is enabled or disabled
         **/
        //IDLOOP = 1;
        //#define initLoopSpecEngine(iteration,ini,IDLOOP)
        
        //iteration = k;
        for (tid = 0; tid <= threads - 1; tid++) {

            ini = 0;
            if (getFirstBlock(&current, &nonSpeculative) == 0) {
                /*printf("NO TIENES TRABAJO %d \n",omp_get_thread_num());*/
                goto labelEndLoop;
            }
            iteration = varblock[0][current] + ini;
            varblock[2][current] = varblock[2][current] + 1;
            //printf("\t\tvarblock[0][current]:%d,current:%d,iteration:%d\n",varblock[0][current],current,iteration);
labelStartIteration: //printf("ENDLOOP,iteration:%d,current:%d\n",iteration,current);
            //  Codigo del bucle
k=iteration;
                            rx = input[0][k];
                            ry = input[1][k];

                           //***********************************************************************
// speccode: speculative load. Original line:
//     spec_centrox = result[centrox];
       linear = centrox;
       if( specload_pointer((unsigned char *) &(result[linear]), sizeof(result[linear]), current, (unsigned char *) &value) == -1)
	  earlySquash(1);
       spec_centrox = value;
//***********************************************************************
//***********************************************************************
// speccode: speculative load. Original line:
//     spec_centroy = result[centroy];
       linear = centroy;
       if( specload_pointer((unsigned char *) &(result[linear]), sizeof(result[linear]), current, (unsigned char *) &value) == -1)
	  earlySquash(1);
       spec_centroy = value;
//***********************************************************************
//***********************************************************************
// speccode: speculative load. Original line:
//     spec_cuad_radio = result[cuad_radio];
       linear = cuad_radio;
       if( specload_pointer((unsigned char *) &(result[linear]), sizeof(result[linear]), current, (unsigned char *) &value) == -1)
	  earlySquash(1);
       spec_cuad_radio = value;
//***********************************************************************
	if (cuad_dist(rx,ry,spec_centrox,spec_centroy) - spec_cuad_radio > EPS){

	 // Inlined version of circunf3p() follows 
	 xpq = px-qx;
	 xqr = qx-rx;
	 xpr = px-rx;
	 ypq = py-qy;
	 yqr = qy-ry;
	 ypr = py-ry;
	 denom = 2*(xpq*ypr-ypq*xpr);

	 if ( denom == 0 ){
	  printf("Error: 3 aligned points.\n");
	  printf("i = %d\n",i);
	  printf("j = %d\n",j);
	  printf("k = %d\n",k);
	  printf("px = %f , py = %f\n",px,py);
	  printf("qx = %f , qy = %f\n",qx,qy);
	  printf("rx = %f , ry = %f\n",rx,ry);
	  exit(-1);
	 }
	 if (py != ry){
	  lambda1 = (yqr*ypr + xpr*xqr) / denom;

//***********************************************************************
// speccode: speculative store (DEFERRED). Original line:
//     result[centrox] = (px+qx)/2 + lambda1*(py-qy);
       spec_centrox = (px+qx)/2 + lambda1*(py-qy);
// speccode: end of speculative store
//***********************************************************************

//***********************************************************************
// speccode: speculative store (DEFERRED). Original line:
//     result[centroy] = (py+qy)/2 + lambda1*(qx-px);
       spec_centroy = (py+qy)/2 + lambda1*(qx-px);
// speccode: end of speculative store
//***********************************************************************

	 }else{
	  if (py != qy){
	   lambda2 = (yqr*ypq+xqr*xpq) / denom;
//***********************************************************************
// speccode: speculative store (DEFERRED). Original line:
//     result[centrox] = (px+rx)/2 + lambda2*(py-ry);
       spec_centrox  = (px+rx)/2 + lambda2*(py-ry);
       // linear = centrox;
       // value = (px+rx)/2 + lambda2*(py-ry);
       // specstore(linear, current, value, &my_IV_tail);
//***********************************************************************
//***********************************************************************
// speccode: speculative store. Original line:
//     result[centroy] = (py+ry)/2 + lambda2*(rx-px);
       spec_centroy  = (py+ry)/2 + lambda2*(rx-px);
// speccode: end of speculative store
//***********************************************************************

	  }
	 }

// speccode: speculative store (DEFERRED)
        spec_cuad_radio = cuad_dist(px, py, spec_centrox, spec_centroy);
// speccode: end of speculative store

// speccode: speculative store (DEFERRED)
        spec_solrx = rx;
// speccode: end of speculative store

// speccode: speculative store (DEFERRED)
        spec_solry = ry;
// speccode: end of speculative store

// speccode: speculative store (DEFERRED)
        spec_puntdef = 3;
// speccode: end of speculative store

// speccode: Before processing next point k, publish changes on speculative data

//***********************************************************************
// speccode: speculative store.
       value = spec_centrox;
       //specstore(centrox, current, value, &my_IV_tail);
	specstore_pointer((unsigned char *) &(result[centrox]), sizeof (result[centrox]), current, (unsigned char *) &value);
// speccode: end of speculative store

//***********************************************************************
// speccode: speculative store.
       value = spec_centroy;
       //specstore(centroy, current, value, &my_IV_tail);
	specstore_pointer((unsigned char *) &(result[centroy]), sizeof (result[centroy]), current, (unsigned char *) &value);
// speccode: end of speculative store
//***********************************************************************

//***********************************************************************
// speccode: speculative store.
       value = spec_cuad_radio;
       //specstore(cuad_radio, current, value, &my_IV_tail);
	specstore_pointer((unsigned char *) &(result[cuad_radio]), sizeof (result[cuad_radio]), current, (unsigned char *) &value);
// speccode: end of speculative store
//***********************************************************************

//***********************************************************************
// speccode: speculative store.
       value = spec_solrx;
       //specstore(solrx, current, value, &my_IV_tail);
	specstore_pointer((unsigned char *) &(result[solrx]), sizeof (result[solrx]), current, (unsigned char *) &value);
// speccode: end of speculative store
//***********************************************************************

//***********************************************************************
// speccode: speculative store.
       value = spec_solry;
       //specstore(solry, current, value, &my_IV_tail);
	specstore_pointer((unsigned char *) &(result[solry]), sizeof (result[solry]), current, (unsigned char *) &value);
// speccode: end of speculative store
//***********************************************************************

//***********************************************************************
// speccode: speculative store.
       value = spec_puntdef;
       //specstore(puntdef, current, value, &my_IV_tail);
	specstore_pointer((unsigned char *) &(result[puntdef]), sizeof (result[puntdef]), current, (unsigned char *) &value);
// speccode: end of speculative store
//***********************************************************************

      }// end if

// endLoopSpecEngine(result,k,j-1,0,1);
            //iteration = k;
            nIterations = j;
            ini = 0;
            /**
             * @author Alvaro Estebanez Lopez
             * @arg iteration index of the iteration is being excuted inside the speculative loop
             * @arg nIterations total number of iterations in the loop
             * @arg ini initial index of the loop
             * @arg IDLOOP loop identifier
             * Returning point after ending an iteration
             * A new block of iterations will be assigned providing that
             * there are still work to execute
             **/
            //#define endLoopSpecEngine(iteration, nIterations, ini, IDLOOP)
            /**********************************************************************		\
            // speccode: end of thread and commit						\
            //           First check if block of iterations is done **************/		\
	if (endLoop || endReturn) {
                freeWheel(current);
                goto labelEndLoop;
            }
labelEndIteration:
            if ((iteration != varblock[1][current] + ini) && (iteration < nIterations - 1)) {
                iteration = iteration + 1;
                goto labelStartIteration;
            }
            /* Thread done, perform the commit */
labelSquash:
            retflag = threadend_pointer(&current, &nonSpeculative);
            /* End if job done */
            if (retflag == JOBDONE) goto labelEndLoop;
            if (retflag == JOBTODO) {
                /*  Set loop variable */
                iteration = varblock[0][current] + ini;
                varblock[2][current] = varblock[2][current] + 1;
                goto labelStartIteration;
            }
labelEndLoop:
            ; /*freeWheel(current);*/
        } /*End for initLoop()*/
        //} //End pragma omp
    } // #pragma omp parallel
    // End of the time from this loop
    stopTime();
    }// end if
   }// end for j
  }// end if
 } // end for i

    //	End of process



    // --------------------------------------------------------------------
    //	Check results

    for (i = 0; i < N; i++) {
        px = input[0][i];
        py = input[1][i];
        if (cuad_dist(px, py, result[centrox], result[centroy]) - result[cuad_radio] > EPS)
            printf("mec-2d: Fail!\n");
    }

    // --------------------------------------------------------------------
    //	Show results

    printf("mec-2d: Points processed: %d\n", N);
    printf("mec-2d: Center of MEC: ('%f', '%f')\n", result[centrox], result[centroy]);
    printf("mec-2d: Squared radius of MEC: %f\n", result[cuad_radio]);
    printf("mec-2d: Number of points that define the solution: %f\n", result[puntdef]);
    if (result[puntdef] >= 1)
        printf("mec-2d: Point one: ('%f', '%f')\n", result[solpx], result[solpy]);
    if (result[puntdef] >= 2)
        printf("mec-2d: Point two: ('%f', '%f')\n", result[solqx], result[solqy]);
    if (result[puntdef] >= 3)
        printf("mec-2d: Point three: ('%f', '%f')\n", result[solrx], result[solry]);
    printf("\n\nmec2d: Timing K loop: %f seconds\n", counter);

    printf("\n\n----SPEC----\n\n");
    //specepilogue();
//    exceptionsEpilogue();
    return (0);
}// main
// =====================================================================

// =====================================================================

double cuad_dist(double px, double py, double qx, double qy) {

    double res1, res2;

    res1 = (px - qx);
    res1 = res1*res1;
    res2 = (py - qy);
    res2 = res2*res2;

    return (res1 + res2);

}

// =====================================================================

